module Navigation (main) where 

import ConfigureMe
import IO


{-
  reads the whole contents of a file
-}
readF :: FilePath -> IO String
readF path = do h <- openFile path ReadMode
		len <- hFileSize h
		s <- hGetContents h
		let dumm = last s
		if (dumm == 'c') then putStr "" else putStr ""
		hClose h
		return s
{-
  writes the whole contents of a file
-}
writeF :: FilePath -> String -> IO ()
writeF path txt = do h <- openFile path WriteMode
		     hPutStr h txt
		     hClose h

-----------------------------------------------------------------
--  Scanner and Parser for AProVETOC.xml
-----------------------------------------------------------------
data Token = Open String String | Close deriving Show

tt_and_close ys =
   let (' ':'t':'e':'x':'t':'=':'"':xs) = ys
       (str,zs) = break ('"' ==) xs
       ('"':' ':'t':'a':'r':'g':'e':'t':'=':'"':us) = zs
       (str2,ws) = break ('"' ==) us
       in if (take 2 ws == "\"/") 
	    then Open str str2:Close:scan ws
            else Open str str2:scan ws
	     

scan :: String -> [Token]
scan ('<':'t':'o':'c':'i':'t':'e':'m':xs) = 
    tt_and_close xs
scan ('<':'/':'t':'o':'c':'i':'t':'e':'m':'>':xs) = 
    Close : scan xs
scan (_:xs) = scan xs
scan "" = []



--------------------------------------------------------------------
-- Parser

data Tree a = Node a [Tree a] 

parse :: [Token] -> ([Token],Tree (String,String))
parse ((Open t1 t2):ts) =
    let (Close:ts',list) = parseTL ts []
	in (ts',Node (t1,t2) list)

parseTL :: [Token] -> [Tree (String,String)] -> ([Token], [Tree (String,String)])
parseTL o@((Open t1 t2):_) list =
    let (ts,tree) = parse o
	in parseTL ts (list++[tree])
parseTL (c@(Close : _)) list 
    = (c, list)
parseTL [] list = 
    ([], list)


------------------------------------------------------
--  parses the TOC file to a Tree representing the toc
--  in the tree are pairs of strings representing title and path
------------------------------------------------------
parseTOC :: String -> Tree (String,String)
parseTOC s = (snd . parse . scan ) s

-- end of parser
-----------------------------------------------------------------


	
-----------------------------------------------------------------
-- translate tree to a tasklist (a mapping from html-file to html-code)
-----------------------------------------------------------------

-- flattens a tree to a list storing each node and its parents in the list
treeToList :: Tree a -> [(Tree a, [a])]
treeToList = treeToList' []
 where treeToList' :: [a] -> Tree a -> [(Tree a,[a])]
       treeToList' parents n@(Node a ns) =
          (n,parents):concat (map (treeToList' (a:parents)) ns)

listToTask :: [(Tree (String,String), [(String,String)])] -> [(String,HTML)]
listToTask ls = lTT (( \ (Node (_,p) _,_ ) -> p) (head ls)) (( \ (Node (_,p) _,_ ) -> p) (last ls)) ls
    where lTT :: String -> String -> [(Tree (String,String), [(String,String)])] -> [(String,String)]
          lTT head left ((Node (titel,path) children, parents):rest) =
	      (path,"<!-- begin navigation -->\n<!-- automatically generated navigation, changes will be overwritten! -->\n" ++ (create path titel left (safeHead head rest) parents (map (\ (Node (t,l) _) -> (t,l)) children))++"<!-- end navigation -->") : lTT head path rest
	  lTT _ _ [] = []
	  safeHead head [] = head
	  safeHead _ ((Node (_,p) _,_) :_) = p
	  

-----------------------------------------------------------------
-- process a single task
-----------------------------------------------------------------

-- looks for a navigation point in a String and delivers a
-- function inserting a string in the navigation point

findNavigation :: String -> Maybe (String -> String)
findNavigation = findNavigation1 ""
    where findNavigation1 fst ('<':'!':'-':'-':' ':'b':'e':'g':'i':'n':' ':'n':'a':'v':'i':'g':'a':'t':'i':'o':'n':' ':'-':'-':'>':rest) = findNavigation2 (reverse fst) rest
	  findNavigation1 fst (x:rest) = findNavigation1 (x:fst) rest
	  findNavigation1 _ [] = Nothing
	  findNavigation2 t ('<':'!':'-':'-':' ':'e':'n':'d':' ':'n':'a':'v':'i':'g':'a':'t':'i':'o':'n':' ':'-':'-':'>':rest) = Just ( \s -> t ++ s ++ rest)
	  findNavigation2 t (_:rest) = findNavigation2 t rest
	  findNavigation2 _ [] = Nothing


-- looks for a body center /center combination as a navigation point
findNoNavigation :: String -> Maybe (String,String)
findNoNavigation = findBody ""
    where findBody fst txt | take 6 (map toUpper txt) == "<BODY>" = 
	    let rest = dropWhile (\x -> elem x " \n\t") (drop 6 txt)
		in if (take 8 (map toUpper rest)) == "<CENTER>" 
		     then findClosingCenter (reverse fst ++ take 6 txt) "" (drop 6 txt)
		     else Nothing
	  findBody fst (x:xs) = findBody (x:fst) xs
	  findBody _ "" = Nothing
	  findClosingCenter fst snd txt | take 9 (map toUpper txt) == "</CENTER>" =
	    let naviBlock = (reverse snd) ++ take 9 txt
		finalText = fst ++ "\n<!-- begin navigation -->\n"++naviBlock++"\n<!-- end navigation -->\n"++drop 9 txt
	       in Just (naviBlock, finalText)
	  findClosingCenter fst snd (x:xs) = findClosingCenter fst (x:snd) xs
	  findClosingCenter _ _ "" = Nothing



processTask :: (String,HTML) -> IO ()
processTask (path,html) =
    do let file = (map ( \x -> if x == '.' then '/' else x ) path) ++ ".html"
       putStrLn ("Processing: "++file)
       input <- readF file
       let navi = findNavigation input
       case navi of 
         Just f -> let newTxt = f html
		      in finally file newTxt
	 Nothing -> do putStr ("Navigation mark not found!\n")
		       let newNavi = findNoNavigation input
		       case newNavi of
		         Just (nav,new) -> 
			     do putStrLn ("Navigation pattern found:\n"++nav)
				answer <- ask
				if answer then finally file new
				          else putStr "Navigation pattern not accepted, handle manually!\n\n"
			 Nothing -> putStr "No Navigation pattern found\nHandle manually!\n\n"
  where finally file txt = do writeF file txt
			      putStr (file++" properly updated\n\n")
	ask = do putStrLn "Is this okay (y/n)?"
		 c <- getChar
		 if (elem c "yY") 
		    then return True
		    else if (elem c "nN") 
		             then return False
			     else ask
		 






main :: IO ()
main = do putStr "Reading TOC..."
	  t <- readF "AProVETOC.xml"
	  putStrLn "okay\nParsing TOC..."
	  let tree = parseTOC t
          let taskList = listToTask (treeToList tree)
	  if (length (show taskList) > 0)
            then do putStrLn "okay\nProcessing Files...\n"
	            mapM processTask taskList
		    return ()
            else putStrLn "Problem in Parsing TOC"



