package benchmarks.stack;

import java.util.ArrayList;

import rwth.i2.ltlrv.LTL;


public class Stack extends ArrayList {
    
    private boolean isFaulty;
    
    private Stack(boolean isFaulty) {
        this.isFaulty = isFaulty;
    }
    
    public void push(Object t) {
        this.add(t);
    }
    
    public Object pop() {
        if(isFaulty && size()>1) {
            return this.remove(this.size()-2);
        }
        return this.remove(this.size()-1);
    }
    
    public Object top() {
        return this.get(this.size()-1);
    }
    
    public boolean isEmpty() {
        return this.size()==0;
    }
    
    public int length() {
        if(isFaulty) {
            return 0;
        }
        return super.size();
    }
    
    public String toString() {
        return "(This stack is "+(isFaulty?"faulty":"correct")+".)";
    }
    
    public static void main(String[] args) {
        
        //perform some tests
        
        Stack s = new Stack(false);

        test(s);
        
        s = new Stack(true);
        
        test(s);

    }

    static void test(Stack s) {
        System.out.println("Push(1)");
        s.push("1");
        System.out.println("Push(2)");
        s.push("2");
        System.out.println("Length()");
        s.length();
        System.out.println("Pop()");
        s.pop();
        System.out.println("Push(3)");
        s.push("3");
        System.out.println("Top()");
        s.top();
        System.out.println("Pop()");
        s.pop();
        System.out.println("Length()");
        s.length();
        System.out.println("Pop()");
        s.pop();
    }
    
    
}

aspect CheckStack {
    
    pointcut push(Stack s, Object o): call(void Stack.push(Object)) && target(s) && args(o);
    pointcut pop(Stack s): call(Object Stack.pop()) && target(s);
    pointcut top(Stack s): call(Object Stack.top()) && target(s);
    
    /*
     * Wishlist: Reusable symbol definitions.
     */
    
    /*
     * TEST 1
     * Checks that a stack returns the last pushed object when pop is invoked. 
     * 
     * Wishlist: Would be great to have context-free expressions here.
     * Does cflowdepth help? -> presumably not
     */    
    tracematch(Stack s, Object pushed, Object popped) {
        
        sym push before: push(s,pushed);
        sym pop after returning (popped): pop(s);
        //this bit was tricky -> it's easily overseen that this is necessary
        sym push2 before: push(s,*); 
        sym pop2 after returning : pop(s);

        //G(push(s,pushed) -> ! (!(push(*,*) || pop(*)) U(pop(s,popped) && if(pushed!=popped))))
        push pop {            
            if(pushed!=popped) {
                System.out.println("Test 1 failed.");
                System.out.println(s);
                System.out.println("Expected: "+pushed+", actual: "+popped);                
            }
        }                
        
    }
    
    /*
     * TEST 2
     * Checks that on a Stack pop() is not invoked before push(Object).
     * 
     * WHY does this fail?
     */    
    tracematch(Stack s) {

        sym startup before: execution(public static void *.main(String[]));
        sym push after: push(s,*);
        sym pop after: pop(s);
        
        //!pop(s) U push(s,*)
        startup pop {
            System.out.println("Test 2 failed.");
            System.out.println(s);
            System.out.println("pop before push!");            
        }
        
    }
    
    /*
     * Test 3
     * Checks that a stack returns the last pushed object when top is invoked
     * (and no pop happened between push and top). 
     * 
     * Wishlist: same as for pop()
     */    
    tracematch(Stack s, Object pushed, Object top) {
        
        sym push before: push(s,pushed);
        sym pop after returning : pop(*);
        sym top after returning (top): top(s);

        //!top(s) U push(s,*)
        push top {            
            if(pushed!=top) {
                System.out.println("Test 3 failed.");
                System.out.println(s);
                System.out.println("Expected: "+pushed+", actual: "+top);                
            }
        }                
        
    }   

    /*
     * Test 4
     * Checks that when something is popped onto a stack,
     * the stack size is not zero.  
     * 
     * Wishlist: with let(..) one could define more properties here
     *           also: if one could bind a variable n in push[n] one could
     *           even track the length
     */     
    tracematch(Stack s) {
        
        sym startup before: execution(public static void *.main(String[]));
        sym push after returning: push(s,*);

        //G!(pop(s) && if(s.length()=0))
        startup push {            
            if(s.length()==0) {
                System.out.println("Test 4 failed.");
                System.out.println(s);
                System.out.println("Expected length()>0, actual length()==0");
            }
        }                
        
    }
        
//    /*
//     * USES let():
//     * Checks that if nothing is popped between two push events, then the size increases
//     * by 1 between both pushes. 
//     */  
//    tracematch(Stack s, int i) {
//        
//        sym push1 after returning: push(s,*) && let(i,s.length());
//        sym push2 after returning: push(s,*);
//        sym pop after returning : pop(*);
//
//        push1 push2 {            
//            if(s.length()!=i+1) {
//                System.out.println("Expected length()=="+(i+1)+", actual length()=="+s.length());
//            }
//        }                
//        
//    }
//    
//    /*
//     * USES let():
//     * Checks that if a pop follows a push then the size before the push is the same as after the pop. 
//     */  
//    tracematch(Stack s, int i) {
//        
//        sym push before: push(s,*) && let(i,s.length());
//        sym pop after returning : pop(*);
//
//        push pop {            
//            if(s.length()!=i+1) {
//                System.out.println("Expected length()=="+(i+1)+", actual length()=="+s.length());
//            }
//        }                
//        
//    }
    
    
}
